/**
 * Created by AA on 7/3/20.
 *
 * This contains the validation for the AWCBL validation operations.
 */

//var models = require('./../models/index')
var Promise = require('promise');
var fs = require('fs');
var path = require('path');
var dateFormat = require('dateformat');
var awcblcommon = require('./awcblcommon');

// The file to be validated must be loaded in the directory below
var validateFileDir = process.env.MASTER_FILE || '/opt/fresenius/efs/sftp/bcdsupport/dropbox/prepare';
var reportsFileDir = process.env.MASTER_FILE || '/opt/fresenius/m3-api/components/public/reports';
var loadingFileDir = process.env.MASTER_FILE || '/opt/fresenius/efs/sftp/bcdsupport/dropbox';

// ======================= CSV Row position declaration ======================================================
//
// Modify this area if there is any change in the input CSV format file
// CAUTION !!!!!  Since the queue format is using these sequences to format the output data for the m3-loader
// on the m3-loader code exists a similar definition of the row position !
// THEY MUST BE KEPT IN SYNC IN CASE OF MODIFICATION !!!!
// See the file component/helpers/LoaderHelper.js on the fresenius-m3-loader code !
// Note that these positions DO NOT include the metadata and also some loader has more data prepared by the
// validation process

// Master Item CSV postion
// row[0]  -> sap_material
// row[1]  -> description
// row[2]  -> cnt_ind
// row[3]  -> cnt_description
// row[4]  -> med_vendor
// row[5]  -> vendor_material
// row[6]  -> conv
// row[7]  -> count_type
// row[8]  -> alternates
// row[9]  -> item_message
// row[10]  -> unit_price
// row[11] -> vendor_id
// row[12] -> onhand_inventory_ind
// row[13] -> discontinued
const rowCsvMTPos = {"sapMaterial":0, "description":1, "cntInd":2, "cntDescription":3, "medVendor":4, "vendorMaterial":5,
  "conv":6, "countType":7, "alternates":8, "itemMessage":9, "unitPrice":10, "vendorId":11,
  "onHand":12, "discontinued":13};

// Clinic Profile CSV position
// row[0]  -> clinic_id
// row[1]  -> vendor_list
// row[2]  -> mircera_visible
// row[3]  -> mircera_io
// row[3]  -> aranesp_visible
// row[3]  -> aranesp_io
// row[3]  -> epogen_visible
// row[3]  -> epogen_io
// row[3]  -> retacrit_visible
// row[3]  -> retacrit_io
// row[3]  -> countin_each
const rowCsvCPPos = {"clinicId":0, "testCatalog":1, "mirceraVisible":2, "mirceraIo":3, "areanespVisible":4, "areanespIo":5,
  "epogenVisible":6, "epogenIo":7, "retacritVisible":8, "retacritIo":9, "countEachIn":10};

// Clinic Routing CSV position
// row[0]  -> clinic_id
// row[1]  -> vendor_id
// row[2]  -> route
const rowCsvCRPos = {"clinicId":0, "vendorId":1, "route":2};

// Par Value CSV position
// row[0]  -> clinic_id
// row[1]  -> product_id
// row[2]  -> qoh
// row[3]  -> minPar
// row[4]  -> maxPar
const rowCsvPVPos = {"clinicId":0, "vendorMaterial":1, "qoh":2, "minPar":3, "maxPar":4};


// ============================================================================================================

// ========>  Validate clinic routing loading file
function ValidateClinicRouting(app, file) {
  return new Promise(function(resolve, reject) {
    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    models.ClinicProfile.findAll().then(function(clinics) {
      if (clinics) {
        logger.info("loaded " + clinics.length + " Clinic profiles");

        models.VendorList.findAll().then(function(vendors) {
          if (vendors) {
            logger.debug("loaded " + vendors.length + " Vendors");

            const results = [];
            const processed_data = [];

            var anyError = false;
            fs.createReadStream(validateFileDir + '/' + file)
                .pipe(csv({
                  headers: false
                }))
                .on('data', (row) => {

                  var errorMessage = "";
                  var status = "Pass";
                  var vendor_display;

                  try {
                    // Verify number of columns per row
                    // This is the first validation test, if is failing is possible other fields are missing !!
                    // Skip the rest of tests if this test fails !
                    errorMessage += verifyColumnCount(row, 3);

                    // Check if there are invalid characters in the row
                    errorMessage += checkInvalidCharacters(row);

                    if (errorMessage.length === 0) {
                      // No errors - proceed

                      // Verify if clinic exists via db
                      errorMessage += validateDbCheckClinicId(clinics, row[rowCsvCRPos.clinicId]);

                      // Verify if vendor ID exists via db
                      errorMessage += validateDbCheckVendorId(vendors, row[rowCsvCRPos.vendorId]);

                      // Validate clinic_id
                      errorMessage += validateClinicId(logger, row[rowCsvCRPos.clinicId]);

                      // Validate route
                      errorMessage += validateRouting(row[rowCsvCRPos.route]);

                      // Validate duplicate within the file
                      errorMessage += validateDuplicateInFile(row, results);

                      // Get vendor_display using vendor_id
                      vendor_display = getVendorDisplay(vendors, row[rowCsvCRPos.vendorId]);
                    }
                  } catch (e) {
                    logger.error('ValidateClinicRouting - Parsing errors : ' + e);
                    errorMessage = e;
                  }

                  if (errorMessage !== "") {
                    status = "Fail";
                    anyError = true;
                  }

                  // Prepare record for report
                  var rowData = {
                    clinic_id: row[rowCsvCRPos.clinicId],
                    vendor_id: row[rowCsvCRPos.vendorId],
                    route: row[rowCsvCRPos.route],
                    status: status,
                    error_desc: errorMessage
                  };

                  // Prepare data to be loaded in the queue
                  var processedRowData =
                      row[rowCsvCRPos.clinicId] + "," +
                      row[rowCsvCRPos.vendorId] + ","  +
                      row[rowCsvCRPos.route] + "," +
                      vendor_display;

                  results.push(rowData);
                  processed_data.push(processedRowData);
                })
                .on('end', () => {
                  const csvWriter = createCsvWriter({
                    path: reportsFileDir + '/' + file,
                    header: [{
                      id: 'clinic_id',
                      title: 'Clinic_Id'
                    },
                      {
                        id: 'vendor_id',
                        title: 'Vendor_Id'
                      },
                      {
                        id: 'route',
                        title: 'Route'
                      },
                      {
                        id: 'status',
                        title: 'Status'
                      },
                      {
                        id: 'error_desc',
                        title: 'Error_Desc'
                      }
                    ]
                  });

                  csvWriter
                      .writeRecords(results)
                      .then(() => console.log('Report File Created.'));


                  if (anyError) {
                    logger.info('CR - There are errors. Skipping Loading File.');
                    resolve("FAIL");
                  } else {
                    logger.info('CSV file successfully processed');
                    var transactionId = getHash(file + Date.now());

                    // Add metadata
                    processed_data.unshift(transactionId + "," +
                        validateFileDir + '/' + file +  "," +
                        processed_data.length);
                    resolve(processed_data);
                  }
                });
          } else {
            logger.error("Loading Vendors table failed !");
          }
        });
      } else {
        logger.error("Loading Clinics table failed !");
      }
    });
  });
}


// ========>  Validate clinic profile loading file
function ValidateClinicProfile(app, file) {
  return new Promise(function(resolve, reject) {
    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var list_catalogs;

    // Need to load the catalog table  for validate data

    models.CatalogList.findAll().then(function(catalogs) {
      if (catalogs) {
        logger.info("loaded " + catalogs.length + " Catalogs");

        const results = [];
        const processed_data = [];

        var anyError = false;
        fs.createReadStream(validateFileDir + '/' + file)
            .pipe(csv({
              headers: false
            }))
            .on('data', (row) => {

              var errorMessage = "";
              var status = "Pass";

              try {
                // Verify number of columns per row
                // This is the first validation test, if is failing is possible other fields are missing !!
                // Skip the rest of tests if this test fails !
                errorMessage += verifyColumnCount(row, 11);

                // Check if there are invalid characters in the row
                errorMessage += checkInvalidCharacters(row);

                if (errorMessage.length === 0) {
                  // No errors - proceed

                  // Verify if catalogs exists via db
                  errorMessage += validateCatalogList(logger, catalogs, row[rowCsvCPPos.testCatalog]);

                  // Validate clinic_id
                  errorMessage += validateClinicId(logger, row[rowCsvCPPos.clinicId]);

                  // Validate yes or no values for flags
                  for (var i = rowCsvCPPos.mirceraVisible; i < rowCsvCPPos.countEachIn + 1; i++) {
                    errorMessage += validateYesNo(row[i]);
                  }

                  // Validate duplicate within the file
                  errorMessage += validateDuplicateInFile(row, results);
                }
              } catch (e) {
                logger.error('ValidateClinicProfile - Parsing errors : ' + e);
                errorMessage = e;
              }

              if (errorMessage !== "") {
                status = "Fail";
                anyError = true;
              }

              // Prepare the list of catalog ID to be sent to the queue
              // Verify if clinic exists via db
              list_catalogs = returnCatalogListAsId(logger, catalogs, row[1]);

              // Prepare record for report
              var rowData = {
                clinic_id: row[rowCsvCPPos.clinicId],
                vendor_list: row[rowCsvCPPos.testCatalog],    // Dont report the converted IDs
                mircera_visible: row[rowCsvCPPos.mirceraVisible],
                mircera_io: row[rowCsvCPPos.mirceraIo],
                aranesp_visible: row[rowCsvCPPos.areanespVisible],
                aranesp_io: row[rowCsvCPPos.areanespIo],
                epogen_visible: row[rowCsvCPPos.epogenVisible],
                epogen_io: row[rowCsvCPPos.epogenIo],
                retacrit_visible: row[rowCsvCPPos.retacritVisible],
                retacrit_io: row[rowCsvCPPos.retacritIo],
                countin_each: row[rowCsvCPPos.countEachIn],
                status: status,
                error_desc: errorMessage
              };

              // Prepare data to be loaded in the queue
              var processedRowData =
                  row[rowCsvCPPos.clinicId] + "," +
                  list_catalogs + "," +
                  row[rowCsvCPPos.mirceraVisible] + ","  +
                  row[rowCsvCPPos.mirceraIo] + "," +
                  row[rowCsvCPPos.areanespVisible] + "," +
                  row[rowCsvCPPos.areanespIo] + "," +
                  row[rowCsvCPPos.epogenVisible] + "," +
                  row[rowCsvCPPos.epogenIo] + "," +
                  row[rowCsvCPPos.retacritVisible] + "," +
                  row[rowCsvCPPos.retacritIo] + "," +
                  row[rowCsvCPPos.countEachIn];

              results.push(rowData);
              processed_data.push(processedRowData);
            })
            .on('end', () => {
              const csvWriter = createCsvWriter({
                path: reportsFileDir + '/' + file,
                header: [{
                  id: 'clinic_id',
                  title: 'Clinic_Id'
                },
                  {
                    id: 'vendor_list',
                    title: 'Vendor_list'
                  },
                  {
                    id: 'mircera_visible',
                    title: 'Mircera_visible'
                  },
                  {
                    id: 'mircera_io',
                    title: 'Mircera_io'
                  },
                  {
                    id: 'aranesp_visible',
                    title: 'Aranesp_visible'
                  },
                  {
                    id: 'aranesp_io',
                    title: 'Aranesp_io'
                  },
                  {
                    id: 'epogen_visible',
                    title: 'Epogen_visible'
                  },
                  {
                    id: 'retacrit_visible',
                    title: 'Retacrit_visible'
                  },
                  {
                    id: 'retacrit_io',
                    title: 'Retacrit_io'
                  },

                  {
                    id: 'countin_each',
                    title: 'Countin_each'
                  },
                  {
                    id: 'status',
                    title: 'Status'
                  },
                  {
                    id: 'error_desc',
                    title: 'Error_Desc'
                  }
                ]
              });

              csvWriter
                  .writeRecords(results)
                  .then(() => console.log('Report File Created.'));

              if (anyError) {
                logger.info('CP - There are errors. Skipping Loading File.');
                resolve("FAIL");
              } else {
                logger.info('CSV file successfully processed');
                var transactionId = getHash(file + Date.now());

                // Add metadata
                processed_data.unshift(transactionId + "," +
                    validateFileDir + '/' + file +  "," +
                    processed_data.length);
                resolve(processed_data);
              }
            });
      } else {
        logger.error("Loading Catalog table failed !");
      }
    });
  });
}

// ======> Validate master item loading file
function ValidateMasterItem(app, file) {
  return new Promise(function(resolve, reject) {

    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var list_catalogs;

    models.VendorList.findAll().then(function (vendors) {
      if (vendors) {
        logger.debug("loaded " + vendors.length + " Vendors");

        models.MasterItem.findAll().then(function (items) {
          if (items) {
            logger.debug("loaded " + items.length + " Items");

            models.CatalogList.findAll().then(function (catalogs) {
              if (catalogs) {
                logger.info("loaded " + catalogs.length + " Catalogs");

                const duplicates = [];
                const results = [];
                const processed_data = [];
                const vendor_materials = [];

                var anyError = false;
                fs.createReadStream(validateFileDir + '/' + file)
                    .pipe(csv({
                      headers: false
                    }))
                    .on('data', (row) => {
                      vendor_materials.push(row[rowCsvMTPos.vendorMaterial]);
                    })
                    .on('end', () => {
                      fs.createReadStream(validateFileDir + '/' + file)
                          .pipe(csv({
                            headers: false
                          }))
                          .on('data', (row) => {

                            var errorMessage = "";
                            var status = "Pass";

                            try {
                              // Validation generic characteristics of input record
                              // Verify number of columns per row  (FRES-192 - removal of NDC field)
                              // (FRES-199 adding item_message field)
                              // This is the first validation test, if is failing is possible other fields are missing !!
                              // Skip the rest of tests if this test fails !
                              errorMessage += verifyColumnCountWithMessage(row, 14);
                              errorMessage += validateifLineIsTitle(row[rowCsvMTPos.sapMaterial], "sapMaterial");

                              // Check if there are invalid characters in the row
                              errorMessage += checkInvalidCharacters(row);

                              if (errorMessage.length === 0) {
                                // No errors - proceed

                                // Verify

                                // Verify if vendor exists via db
                                errorMessage += validateDbCheckVendorId(vendors, row[rowCsvMTPos.vendorId]);

                                // Verify if discontinued date exists
                                errorMessage += validateValidDate(row[rowCsvMTPos.discontinued]);

                                // Verify alternate
                                if (row[rowCsvMTPos.alternates] !== "") {
                                  errorMessage += validateAlternate(items, vendor_materials, row[rowCsvMTPos.alternates]);
                                }

                                // Verify if catalogs exists via db
                                errorMessage += validateCatalogList(logger, catalogs, row[rowCsvMTPos.medVendor]);

                                // Verify / Correct - resolve quotation marks
                                row[rowCsvMTPos.itemMessage] = validateCorrectMessage(row[rowCsvMTPos.itemMessage]);

                                // Validate boolean entries
                                logger.debug("  ---> validateX on : " + row[rowCsvMTPos.cntInd]);
                                errorMessage += validateX(row[rowCsvMTPos.cntInd]);
                                logger.debug("  ---> validateX on : " + row[rowCsvMTPos.onHand]);
                                errorMessage += validateX(row[rowCsvMTPos.onHand]);

                                // Validate duplicate within the file
                                errorMessage += validateDuplicateInFileMT(row, duplicates);
                              }
                            } catch (e) {
                              logger.error('ValidateMasterItem - Parsing errors : ' + e);
                              errorMessage = 'ValidateMasterItem - Parsing errors : ' + e;
                            }

                            if (errorMessage !== "") {
                              status = "Fail";
                              anyError = true;
                            }

                            // Prepare the list of catalog ID to be sent to the queue
                            // If errors exists for the catalog should be already handled by the validation above
                            list_catalogs = returnCatalogListAsId(logger, catalogs, row[rowCsvMTPos.medVendor]);

                            var rowData = {
                              sap_material: row[rowCsvMTPos.sapMaterial],
                              description: row[rowCsvMTPos.description],
                              cnt_ind: row[rowCsvMTPos.cntInd],
                              cnt_description: row[rowCsvMTPos.cntDescription],
                              med_vendor: row[rowCsvMTPos.medVendor],
                              vendor_material: row[rowCsvMTPos.vendorMaterial],
                              conv: row[rowCsvMTPos.conv],
                              count_type: row[rowCsvMTPos.countType],
                              alternates: row[rowCsvMTPos.alternates],
                              item_message: row[rowCsvMTPos.itemMessage],
                              unit_price: row[rowCsvMTPos.unitPrice],
                              vendor_id: row[rowCsvMTPos.vendorId],
                              onhand_inventory_ind: row[rowCsvMTPos.onHand],
                              discontinued: row[rowCsvMTPos.discontinued],

                              status: status,
                              error_desc: errorMessage
                            };

                            // Prepare data to be loaded in the queue
                            var processedRowData =
                                row[rowCsvMTPos.sapMaterial] + "," +
                                "\"" + row[rowCsvMTPos.description] + "\"," +
                                row[rowCsvMTPos.cntInd] + "," +
                                row[rowCsvMTPos.cntDescription] + "," +
                                list_catalogs + "," +
                                row[rowCsvMTPos.vendorMaterial] + "," +
                                row[rowCsvMTPos.conv] + "," +
                                row[rowCsvMTPos.countType] + "," +
                                row[rowCsvMTPos.alternates] + "," +
                                "\"" + row[rowCsvMTPos.itemMessage] + "\"," +
                                row[rowCsvMTPos.unitPrice] + "," +
                                row[rowCsvMTPos.vendorId] + "," +
                                row[rowCsvMTPos.onHand] + "," +
                                row[rowCsvMTPos.discontinued];

                            results.push(rowData);
                            processed_data.push(processedRowData);
                          })
                          .on('end', () => {
                            const csvWriter = createCsvWriter({
                              path: reportsFileDir + '/' + file,
                              header: [{
                                id: 'sap_material',
                                title: 'Sap_Material'
                              },
                                {
                                  id: 'description',
                                  title: 'Description'
                                },
                                {
                                  id: 'cnt_ind',
                                  title: 'Cnt_Ind'
                                },
                                {
                                  id: 'cnt_description',
                                  title: 'Cnt_Description'
                                },
                                {
                                  id: 'med_vendor',
                                  title: 'Med_Vendor'
                                },
                                {
                                  id: 'vendor_material',
                                  title: 'Vendor_material'
                                },
                                {
                                  id: 'conv',
                                  title: 'Conv'
                                },
                                {
                                  id: 'count_type',
                                  title: 'Count_Type'
                                },
                                {
                                  id: 'alternates',
                                  title: 'Alternates'
                                },
                                {
                                  id: 'item_message',
                                  title: 'Item_Message'
                                },
                                {
                                  id: 'unit_price',
                                  title: 'Unit_Price'
                                },
                                {
                                  id: 'vendor_id',
                                  title: 'Vendor_Id'
                                },
                                {
                                  id: 'onhand_inventory_ind',
                                  title: 'Onhand_inventory_ind'
                                },
                                {
                                  id: 'discontinued',
                                  title: 'Discontinued'
                                },
                                {
                                  id: 'status',
                                  title: 'Status'
                                },
                                {
                                  id: 'error_desc',
                                  title: 'Error_Desc'
                                }
                              ]
                            });

                            // Write the report file
                            csvWriter
                                .writeRecords(results)
                                .then(() => console.log('Report File Created.'));

                            // If no errors send prepared load to the loader
                            if (anyError) {
                              logger.info('MT - There are errors. Skipping Loading File.');
                              resolve("FAIL");
                            } else {
                              logger.info('CSV file successfully processed');
                              var transactionId = getHash(file + Date.now());

                              // Add metadata
                              processed_data.unshift(transactionId + "," +
                                  validateFileDir + '/' + file +  "," +
                                  processed_data.length);

                              resolve(processed_data);
                            }
                          });
                    });
              } else {
                logger.error("Loading Catalog table failed !");
              }
            });
          } else {
            logger.error("Loading Master Items table failed !");
          }
        });
      } else {
        logger.error("Loading Vendor List table failed !");
      }
    });
  });
}

// ========>  Validate Par value loading file
function ValidateParValue(app, file) {
  return new Promise(function(resolve, reject) {

    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var vendor_id;
    var material_description;
    var vendor_display;

    models.ClinicProfile.findAll().then(function(clinics) {
      if (clinics) {
        logger.info("loaded " + clinics.length + " Clinic profiles");

        models.MasterItem.findAll().then(function(items) {
          if (items) {
            logger.debug("loaded " + items.length + " Items");

            models.VendorList.findAll().then(function (vendors) {
              if (vendors) {
                logger.debug("loaded " + vendors.length + " Vendors");

                const results = [];
                const processed_data = [];

                var anyError = false;
                fs.createReadStream(validateFileDir + '/' + file)
                    .pipe(csv({
                      headers: false
                    }))
                    .on('data', (row) => {

                      var errorMessage = "";
                      var status = "Pass";

                      try {
                        // Verify number of columns per row
                        // This is the first validation test, if is failing is possible other fields are missing !!
                        // Skip the rest of tests if this test fails !
                        errorMessage += verifyColumnCount(row, 5);

                        // Check if there are invalid characters in the row
                        errorMessage += checkInvalidCharacters(row);

                        if (errorMessage.length === 0) {
                          // No errors - proceed

                          // Verify if clinic exists via db
                          errorMessage += validateDbCheckClinicId(clinics, row[rowCsvPVPos.clinicId]);

                          // Verify if vendor material exists via db
                          errorMessage += validateVendorMaterial(items, row[rowCsvPVPos.vendorMaterial]);

                          // Validate clinic_id
                          errorMessage += validateClinicId(logger, row[rowCsvPVPos.clinicId]);

                          // Validate duplicate within the file
                          errorMessage += validateDuplicateInFile(row, results);

                          // Get vendor_id using vendor_material
                          vendor_id = getVendorId(items, row[rowCsvPVPos.vendorMaterial]);

                          // Get vendor_display using vendor_id
                          vendor_display = getVendorDisplay(vendors, vendor_id);

                          // Get material_description using vendor_material
                          material_description = getMaterialDesc(items, row[rowCsvPVPos.vendorMaterial]);
                        }
                      } catch (e) {
                        logger.error('ValidateParValue - Parsing errors : ' + e);
                        errorMessage = e;
                      }

                      if (errorMessage !== "") {
                        status = "Fail";
                        anyError = true;
                      }

                      // Prepare record for report
                      rowData = {
                        clinic_id: row[rowCsvPVPos.clinicId],
                        vendor_material: row[rowCsvPVPos.vendorMaterial],
                        qoh: row[rowCsvPVPos.qoh],
                        min_par: row[rowCsvPVPos.minPar],
                        max_par: row[rowCsvPVPos.maxPar],
                        status: status,
                        error_desc: errorMessage
                      };

                      // Prepare data to be loaded in the queue
                      var processedRowData =
                          row[rowCsvPVPos.clinicId] + "," +
                          row[rowCsvPVPos.vendorMaterial] + "," +
                          row[rowCsvPVPos.qoh] + "," +
                          row[rowCsvPVPos.minPar] + "," +
                          row[rowCsvPVPos.maxPar] + "," +
                          vendor_id + "," +
                          vendor_display + "," +
                          material_description;

                      results.push(rowData);
                      processed_data.push(processedRowData);

                    })
                    .on('end', () => {
                      const csvWriter = createCsvWriter({
                        path: reportsFileDir + '/' + file,
                        header: [{
                          id: 'clinic_id',
                          title: 'Clinic_Id'
                        },
                          {
                            id: 'vendor_material',
                            title: 'Vendor_Material'
                          },
                          {
                            id: 'qoh',
                            title: 'Qoh'
                          },
                          {
                            id: 'min_par',
                            title: 'Min_Par'
                          },
                          {
                            id: 'max_par',
                            title: 'Max_Par'
                          },
                          {
                            id: 'status',
                            title: 'Status'
                          },
                          {
                            id: 'error_desc',
                            title: 'Error_Desc'
                          }
                        ]
                      });

                      csvWriter
                          .writeRecords(results)
                          .then(() => console.log('Report File Created.'));


                      if (anyError) {
                        logger.info('PV - There are errors. Skipping Loading File.');
                        resolve("FAIL");
                      } else {
                        logger.info('CSV file successfully processed');
                        var transactionId = getHash(file + Date.now());
                        processed_data.unshift(transactionId + "," + validateFileDir + '/' + file + "," + processed_data.length);
                        resolve(processed_data);
                      }
                    });
              } else {
                logger.error("Loading Vendors table failed !");
              }
            });
          } else {
            logger.error("Loading Master Item table failed !");
          }
        });
      } else {
        logger.error("Loading Clinics table failed !");
      }
    });
  });
}

// Validate if the line is a title and not actual data.
// Screen out files generated by the Export function and not edited.
function validateifLineIsTitle(data, title) {
  if (data.includes(title)) {
    return "Line is a title. ";
  } else {
    return "";
  }
}

function validateRouting(data) {
  if (data !== 'IO' && data !== 'SAP') {
    return "Invalid Routing: " + data + ". ";
  } else {
    return "";
  }
}

function validateYesNo(data) {
  if (data !== 'y' && data !== 'n') {
    return "Invalid y/n value: " + data + ". ";
  } else {
    return "";
  }
}

function validateX(data) {
  if (data.length > 1 || (data.length === 1 && data !== 'X')) {
    return "Expected X or nothing: " + data + ". ";
  } else {
    return "";
  }
}

function validateDbCheckClinicId(clinics, clinic_id) {
  var id = 1;
  id = awcblcommon.getClinicId(clinics, clinic_id);
  if (id < 0) {
    //return "Clinic_Id: " + clinic_id + " is not present in the db. id =  " + id + ". ";
    return "Clinic_Id: " + clinic_id + " is not present in the db. ";
  } else {
    return "";
    //return "Clinic_Id: " + clinic_id + " is  present in the db.  id =  " + id + ". ";
  }
}


function validateDbCheckVendorId(vendors, vendor_id) {
  var id = 1;
  id = awcblcommon.getVendorId(vendors, vendor_id);
  if (id < 0) {
    //return "Vendor_Id: " + vendor_id + " is not present in the db.  id =  " + id + ". ";
    return "Vendor_Id: " + vendor_id + " is not present in the db. ";
  } else {
    return "";
    //return "Vendor_Id: " + vendor_id + " is  present in the db.  id =  " + id + ". ";
  }
}

function validateVendorMaterial(allMasterItems, vendMat) {
  var id = 1;
  id = awcblcommon.getProductId(allMasterItems, vendMat);
  if (id < 0) {
    //return "Vendor Material: " + vendMat + " is not present in the db.  id =  " + id + ". ";
    return "Vendor Material: " + vendMat + " is not present in the db. ";
  } else {
    return "";
    //return "Vendor Material: " + vendMat + " is  present in the db.  id =  " + id + ". ";
  }
}

// Alternate validation
// The validate code must exists in the table master_item and be an active, non discontinued product.
function validateAlternate(allMasterItems, vendor_materials, alternate) {
  var found = false;

  // Verify the validate code exists and is a valid record
  for (var v = 0; v < allMasterItems.length; v++) {
    var record = allMasterItems[v];
    if (record.vendorMaterial === alternate && !record.discontinued) {
      found = true;
      break;
    }
  }

  if (found) {
    return "";
  } else {
    return "Invalid alternate: " + alternate + ". ";
  }
}

function validateClinicId(logger, data) {
  // Pattern regexp :
  //
//  var rePattern = new RegExp(/^\.([0-9]+)(BM[0-9]*)$|(HP$)/);
  var rePattern = new RegExp(/^\.([0-9]+)(BM[0-9]*)$|(FM[0-9]*)$|(HP$)|(hp$)|(LIB$)|(HS$)/);

  //  logger.debug('validateClinicId on : ' + data.toString());

  try {
    var result = rePattern.test(data.toString());

    if (result === false) {
      return "Invalid clinic_id: " + data.toString() + " - wrong pattern";
    } else {
      if (data[0] !== '.') {
        return "Invalid clinic_id: " + data.toString() + " - not starting with dot";
      } else {
        return "";
      }
    }
  } catch (err) {
    return "Invalid clinic_id: " + data.toString() + " - undefined error";
  }
}

// The function analyze an input string containing a list of catalogs
// and return an error, empty otherwise
function validateCatalogList(logger, catalogList, data) {
  var returnResult = "";
  var isDefaultPresent = false; // Indicates if a default catalog exists in the data
  var isDefaultExist = false;   // Indicates if the tbl_catalog contains a default catalog
  var strDefCatalogs = "";      // String containing the default catalog entry ready to add
  var testCatalog = "";
  var fileVendorNames = "";
  var cnt = 0;

  // logger.debug("validateCatalogList ");

  try {
    testCatalog = data.split('"').join('');
    fileVendorNames = testCatalog.split(",");
  } catch (e) {
    logger.error("validateCatalogList --> Unable to process incoming data: " + data + " (Error : " + e + ")");
    return "Wrong input : " + e;
  }

  if (!testCatalog) {
    testCatalog = "";
  }

  // Assign default catalogs
  strDefCatalogs = awcblcommon.getDefaultCatalog(catalogList);
  if (strDefCatalogs !== "" && strDefCatalogs !== null) {
    // A default catalog exists !
    isDefaultExist = true;
  }

  if (testCatalog !== "") {
    // catalogs in the input name - verify exists and convert in index

    for (var fn = 0; fn < fileVendorNames.length; fn++) {
      var id = awcblcommon.getCatalogId(catalogList, fileVendorNames[fn].trim());
      if (id >= 0) {
        // Found an ID !
        // logger.debug("   --> Found catalog [" + fileVendorNames[fn].trim() + "] with index " + id);

        // Check to verify if is the default catalog - set a flag to know if to add the default
        // catalog to the returning list
        if (isDefaultExist === true && (id.toString() === strDefCatalogs.substr(1))) {
          // logger.debug("   --> Is the default catalog !");
          isDefaultPresent = true;
        }

        // Don't use the comma as separator !!  Uses #
        returnResult = returnResult + ((cnt > 0) ? "#" : "") + id;
        cnt++;
      } else {
        // No valid ID found !
        logger.error("validateCatalogList --> Catalog parsing error on [" + fileVendorNames[fn] + "] ");
        return "Catalog parsing error on [" + fileVendorNames[fn] + "] ";
      }
    }

    // If the default catalog is NOT present, add it to the list
    if (isDefaultPresent === false && isDefaultExist === true) {
      logger.debug("   --> Add Default catalog to the list");
      returnResult += strDefCatalogs;
    } else {
      logger.debug("   --> No default catalog to be added ");
    }
  } else {
    // No catalogs found - add at least the default one
    // CAUTION !!! Is possible to end up with an empty string if no default catalogs are assigned in the table !
    returnResult = strDefCatalogs.substr(1);
  }

  logger.debug('   --> returning result : ' + returnResult);
  if (returnResult === "" || returnResult === null) {
    return "No catalogs added";
  }

  return "";
}

// The function convert the catalog list in IDs
// It return a string containing the list of the catalog ID present in the data
// or at least a default catalog
// Return the value "" in case of errors (empty list)
function returnCatalogListAsId(logger, catalogList, data) {
  var returnResult = "";
  var isDefaultPresent = false; // Indicates if a default catalog exists in the data
  var isDefaultExist = false;   // Indicates if the tbl_catalog contains a default catalog
  var strDefCatalogs = "";      // String containing the default catalog entry ready to add
  var testCatalog = "";
  var fileVendorNames = "";
  var cnt = 0;

  logger.debug("returnCatalogListAsId ");

  try {
    testCatalog = data.split('"').join('');
    fileVendorNames = testCatalog.split(",");
  } catch (e) {
    logger.error("   --> Unable to process incoming data: " + data + " (Error : " + e + ")");
    return "";
  }

  if (!testCatalog) {
    testCatalog = "";
  }

  // Assign default catalogs
  strDefCatalogs = awcblcommon.getDefaultCatalog(catalogList);
  if (strDefCatalogs !== "" && strDefCatalogs !== null) {
    // A default catalog exists !
    isDefaultExist = true;
  }

  if (testCatalog !== "") {
    // catalogs in the input name - verify exists and convert in index

    for (var fn = 0; fn < fileVendorNames.length; fn++) {
      var id = awcblcommon.getCatalogId(catalogList, fileVendorNames[fn].trim());
      if (id >= 0) {
        // Found an ID !
        //logger.debug("   --> Found catalog [" + fileVendorNames[fn].trim() + "] with index " + id);

        // Check to verify if is the default catalog - set a flag to know if to add the default
        // catalog to the returning list
        if (isDefaultExist === true && (id.toString() === strDefCatalogs.substr(1))) {
          //logger.debug("   --> Is the default catalog !");
          isDefaultPresent = true;
        }

        // Don't use the comma as separator !!  Uses #
        returnResult = returnResult + ((cnt > 0) ? "#" : "") + id;
        cnt++;
      } else {
        // No valid ID found !
        // The error is handled anyway in the validation check - do nothing
        logger.error("   --> Catalog parsing error on [" + fileVendorNames[fn] + "] ");
      }
    }

    // If the default catalog is NOT present, add it to the list
    if (isDefaultPresent === false && isDefaultExist === true) {
      //logger.debug("   --> Add Default catalog to the list");
      returnResult += strDefCatalogs;
    } else {
      //logger.debug("   --> No default catalog to be added ");
    }
  } else {
    // No catalogs found - add at least the default one
    // CAUTION !!! Is possible to end up with an empty string if no default catalogs are assigned in the table !
    returnResult = strDefCatalogs.substr(1);
  }

  logger.debug('   --> returning result : ' + returnResult);
  return returnResult;
}


function validateValidDate(data) {
  if (data === "") {
    return "";
  } else {
    var value = isValidDate(data);
    if (value) {
      return "";
    } else {
      return "Invalid discontinued date: " + data + ". ";
    }
  }
}

function isValidDate(date) {
  try {
    var dateWrapper = new Date(date);
    return dateWrapper && Object.prototype.toString.call(dateWrapper) === "[object Date]" && !isNaN(dateWrapper);
  } catch (err) {
    return false;
  }
}

function validateDuplicateInFile(row, results) {
  var errorMessage = "";

  for (var i = 0; i < results.length; i++) {
    if (row[0] === results[i]["clinic_id"]) {
      if (row[1] === results[i]["vendor_id"]) {
        errorMessage = "Duplicate data in the file: " + row[0] + "," + row[1] + ". ";
        break;
      }
    }
  }

  return errorMessage;
}

function validateDuplicateInFileMT(row, results) {
  var errorMessage = "";
  var previousRows = "";
  var currentRows = row[rowCsvMTPos.sapMaterial] + row[rowCsvMTPos.vendorMaterial];

  if (results.isEmpty) {
    results.push(currentRows);
  } else {
    for (var i = 0; i < results.length; i++) {
      previousRows = results[i];

      if (currentRows === previousRows) {
          errorMessage = "Duplicate data in the file: " + row[rowCsvMTPos.sapMaterial] + "," + row[rowCsvMTPos.vendorMaterial] + ". ";
          break;
      }
    }

    results.push(currentRows);
  }

  return errorMessage;
}


function RemoveFileFromValidationArea(app, file) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');
  var transactionLogger = winston.loggers.get('transaction');

  try {
    fs.unlinkSync(filePath)
    logger.info("Removed File: " + validateFileDir + '/' + file);
    //file removed
  } catch (err) {
    console.error(err)
  }
}


function validateCorrectMessage(row) {
  if(row.length > 0){
    row = row.replace(/"/g, "'");
    row = row.replace(/&quot;/g,"'");
  }
  return row;
}


function verifyColumnCount(obj, expectedCount) {
  if (Object.keys(obj).length !== expectedCount) {
    return "Invalid column count: " + Object.keys(obj).length + ", expected: " + expectedCount + ". ";
  } else {
    return "";
  }
}

function verifyColumnCountWithMessage(obj, expectedCount) {
  if (Object.keys(obj).length !== expectedCount) {
    return "Invalid column count: " + Object.keys(obj).length + ", expected: " + expectedCount + ". Make sure Item_Message is enclosed with Double Quotes.";
  } else {
    return "";
  }
}

function checkInvalidCharacters(row) {

  var rePattern = new RegExp(/^[a-zA-Z0-9 !@#$%^&*()_+\-=\[\]{};':"\\|,.<>\/?]*$/);

  try {
    var result = rePattern.test(JSON.stringify(row));

    if (result === false) {
      return "Invalid character in the row.";
    } else{
      return "";
    }

  } catch (err) {
    return "Invalid character in the row.";
  }
}

// The function looks if in a field under test is present a comma
function validatePresenceComma(name, row) {
  if (row.indexOf(',') > -1) {
    return "Not allowed a comma in the " + name + ".";
  } else{
    return "";
  }
}

function getVendorId(allMasterItems, vendor_material) {
  for (var i = 0; i < allMasterItems.length; i++) {
    var record = allMasterItems[i];
    if (record.vendorMaterial === vendor_material) {
      return record.vendorId;
    }
  }
}

function getMaterialDesc(allMasterItems, vendor_material) {
  for (var i = 0; i < allMasterItems.length; i++) {
    var record = allMasterItems[i];
    if (record.vendorMaterial === vendor_material) {
      return record.description;
    }
  }
}

function getVendorDisplay(vendors, vendor_id) {
  for (var i = 0; i < vendors.length; i++) {
    var record = vendors[i];
    if (record.vendorId === vendor_id) {
      return record.vendorDisplay;
    }
  }
}

function getHash(input){
  var hash = 0, len = input.length;
  for (var i = 0; i < len; i++) {
    hash  = ((hash << 5) - hash) + input.charCodeAt(i);
    hash |= 0; // to 32bit integer
  }
  return hash;
}

module.exports.RemoveFileFromValidationArea = RemoveFileFromValidationArea;
module.exports.ValidateClinicRouting = ValidateClinicRouting;
module.exports.ValidateClinicProfile = ValidateClinicProfile;
module.exports.ValidateParValue = ValidateParValue;
module.exports.ValidateMasterItem = ValidateMasterItem;